/*
 * Copyright (c) 2019-2022, NVIDIA CORPORATION. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <gst/gst.h>
#include <glib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <sys/time.h>
#include <cuda_runtime_api.h>

#include "gst/gstinfo.h"
#include "gstnvdsmeta.h"
#include "nvbufsurface.h"
//#include "gstnvstreammeta.h"
#ifndef PLATFORM_TEGRA
#include "gst-nvmessage.h"
#endif

#define MAX_DISPLAY_LEN 64

#define PGIE_CLASS_ID_VEHICLE 0
#define PGIE_CLASS_ID_PERSON 2

/* The muxer output resolution must be set if the input streams will be of
 * different resolution. The muxer will scale all the input frames to this
 * resolution. */
#define MUXER_OUTPUT_WIDTH 352
#define MUXER_OUTPUT_HEIGHT 416

/* Muxer batch formation timeout, for e.g. 40 millisec. Should ideally be set
 * based on the fastest source's framerate. */
#define MUXER_BATCH_TIMEOUT_USEC 40000

#define TILED_OUTPUT_WIDTH 352
#define TILED_OUTPUT_HEIGHT 416

/* NVIDIA Decoder source pad memory feature. This feature signifies that source
 * pads having this capability will push GstBuffers containing cuda buffers. */
#define GST_CAPS_FEATURES_NVMM "memory:NVMM"

gchar pgie_classes_str[4][32] = { "Vehicle", "TwoWheeler", "Person",
  "RoadSign"
};

#define FPS_PRINT_INTERVAL 300
//static struct timeval start_time = { };

//static guint probe_counter = 0;

/* tiler_sink_pad_buffer_probe  will extract metadata received on OSD sink pad
 * and update params for drawing rectangle, object information etc. */

static GstPadProbeReturn
tiler_src_pad_buffer_probe (GstPad * pad, GstPadProbeInfo * info,
    gpointer u_data)
{
    GstBuffer *buf = (GstBuffer *) info->data;
    guint num_rects = 0; 
    NvDsObjectMeta *obj_meta = NULL;
    guint vehicle_count = 0;
    guint person_count = 0;
    NvDsMetaList * l_frame = NULL;
    NvDsMetaList * l_obj = NULL;
    //NvDsDisplayMeta *display_meta = NULL;
    gchar bbox_file[1024] = { 0 };
    FILE *bbox_params_dump_file = NULL;
    
    NvDsBatchMeta *batch_meta = gst_buffer_get_nvds_batch_meta (buf);

    for (l_frame = batch_meta->frame_meta_list; l_frame != NULL;
      l_frame = l_frame->next) {
        NvDsFrameMeta *frame_meta = (NvDsFrameMeta *) (l_frame->data);
        guint stream_id = frame_meta->pad_index;
        /*g_snprintf (bbox_file, sizeof (bbox_file) - 1,
            "%s/%02u_%03u_%0lu.txt", "./streamscl",
            0, stream_id, (gulong) frame_meta->frame_num);*/
        g_snprintf (bbox_file, sizeof (bbox_file) - 1,
            "%s/%lu.txt", "./streamscl", (gulong) frame_meta->frame_num);
        bbox_params_dump_file = fopen (bbox_file, "w");
        //int offset = 0;
        for (l_obj = frame_meta->obj_meta_list; l_obj != NULL;
                l_obj = l_obj->next) {
            obj_meta = (NvDsObjectMeta *) (l_obj->data);

            // if(obj_meta->obj_label[0] == 'B') {
            //   // badbean_count += 1;
            //   // nvds_remove_obj_meta_from_frame(frame_meta, obj);
            //   nvds_remove_obj_meta_from_frame(batch_meta->frame_meta_list->data, obj_meta);
            //   // nvds_remove_
            //   printf("\nremove bad coffee bean");
            // }

            float left = (obj_meta->rect_params.left);// / frame_meta->pipeline_width;
            float top = (obj_meta->rect_params.top);// / frame_meta->pipeline_height;
            float right = (left + obj_meta->rect_params.width);// / frame_meta->pipeline_width;
            float bottom = (top + obj_meta->rect_params.height);// / frame_meta->pipeline_height;
            // Here confidence stores detection confidence, since dump gie output
            // is before tracker plugin
            float confidence = obj_meta->confidence;
            // printf("\n Ryan \n");
            
            printf ("%s  %f %f %f %f  %f\n",
                obj_meta->obj_label, left, top, right, bottom, confidence);
            //char* label_class = obj_meta->obj_label;
            printf ("\n label:  %c", obj_meta->obj_label[0]);
            printf ("\n frame width %d ", frame_meta->source_frame_width);
            printf ("\n pipeline width %d ", frame_meta->pipeline_width);
            printf ("\n rect_params height %d ", obj_meta->rect_params.height);
            printf ("\n base_meta:  %d", frame_meta->base_meta);
            printf ("\n pad_index:  %d", frame_meta->pad_index);
            
            // fprintf (bbox_params_dump_file, "%c %f %f %f %f %f\n",
            //    obj_meta->obj_label[0], confidence, left, top, right, bottom);
            float x = ((left + right) / 2) / MUXER_OUTPUT_WIDTH;
            float y = ((top + bottom) / 2) / MUXER_OUTPUT_HEIGHT;
            float w = (right - left + 1) / MUXER_OUTPUT_WIDTH;
            float h = (bottom - top + 1) / MUXER_OUTPUT_HEIGHT;

	  //   if (batch_meta->frame_meta_list->data->obj_meta_list->data == NULL)
		// break;
            fprintf (bbox_params_dump_file, "%c %f %f %f %f %f\n",
                obj_meta->obj_label[0], confidence, x, y, w, h);
            // fprintf (bbox_params_dump_file, "%c %f %f %f %f\n",
            //     obj_meta->obj_label[0], x, y, w, h);

            if (obj_meta->class_id == PGIE_CLASS_ID_VEHICLE) {
                vehicle_count++;
                num_rects++;
            }
            if (obj_meta->class_id == PGIE_CLASS_ID_PERSON) {
                person_count++;
                num_rects++;
            }
        }
        
        fclose (bbox_params_dump_file);
        /*g_print ("Frame Number = %d Number of objects = %d "
          "Vehicle Count = %d Person Count = %d\n",
          frame_meta->frame_num, num_rects, vehicle_count, person_count);*/
#if 0
        display_meta = nvds_acquire_display_meta_from_pool(batch_meta);
        NvOSD_TextParams *txt_params  = &display_meta->text_params;
        txt_params->display_text = g_malloc0 (MAX_DISPLAY_LEN);
        offset = snprintf(txt_params->display_text, MAX_DISPLAY_LEN, "Person = %d ", person_count);
        offset = snprintf(txt_params->display_text + offset , MAX_DISPLAY_LEN, "Vehicle = %d ", vehicle_count);

        /* Now set the offsets where the string should appear */
        txt_params->x_offset = 10;
        txt_params->y_offset = 12;

        /* Font , font-color and font-size */
        txt_params->font_params.font_name = "Serif";
        txt_params->font_params.font_size = 10;
        txt_params->font_params.font_color.red = 1.0;
        txt_params->font_params.font_color.green = 1.0;
        txt_params->font_params.font_color.blue = 1.0;
        txt_params->font_params.font_color.alpha = 1.0;

        /* Text background color */
        txt_params->set_bg_clr = 1;
        txt_params->text_bg_clr.red = 0.0;
        txt_params->text_bg_clr.green = 0.0;
        txt_params->text_bg_clr.blue = 0.0;
        txt_params->text_bg_clr.alpha = 1.0;

        nvds_add_display_meta_to_frame(frame_meta, display_meta);
#endif

    }
    return GST_PAD_PROBE_OK;
}

static guint dec_probe_counter = 0;
static guint nvvidcov_probe_counter = 0;
static guint streammux_probe_counter = 0;
#define DUMP_FRAME_CNT_START   (10)
#define DUMP_FRAME_CNT_STOP    (12)

static void dumpNvBufToFile(char* tag, char* suffix, int count, NvBufSurface *surface) {
  FILE *f = NULL;
  if (surface->surfaceList[0].colorFormat == NVBUF_COLOR_FORMAT_NV12)
    printf("it's NVBUF_COLOR_FORMAT_NV12!\n");
  else
    printf("it's not NVBUF_COLOR_FORMAT_NV12(%d)\n", NVBUF_COLOR_FORMAT_NV12);

  int width = surface->surfaceList[0].width;
  int height = surface->surfaceList[0].height;
  int pitch = surface->surfaceList[0].pitch;
  int plane = surface->surfaceList[0].planeParams.num_planes;
  int batchSize = surface->batchSize;
  for (int bs = 0; bs < batchSize; bs++) {
    char fileName[64] = { 0 };
    snprintf(fileName, sizeof(fileName), "bs-%d-%d-%dx%d-%s", bs, count, pitch, height, suffix);
    printf("%s dump to file %s \n", tag, fileName);
    f = fopen(fileName, "wb");
    if (f) {
      for (int i = 0; i < surface->surfaceList[bs].planeParams.num_planes; i++) {
        NvBufSurfaceMap(surface, bs, i, NVBUF_MAP_READ);
        if (surface->memType == NVBUF_MEM_SURFACE_ARRAY) {
          /* Cache the mapped data for CPU access */
          NvBufSurfaceSyncForCpu(surface, bs, i);
        }
        int bpp = surface->surfaceList[bs].planeParams.bytesPerPix[i];
        size_t bytes_to_write = bpp * surface->surfaceList[bs].planeParams.pitch[i] * surface->surfaceList[bs].planeParams.height[i];
        printf("%s plane %d bytes_to_write pitch(%d:bpp %d) x height(%d) = %d \n", tag, i,
          surface->surfaceList[bs].planeParams.pitch[i],
          bpp,
          surface->surfaceList[bs].planeParams.height[i],
          bpp * surface->surfaceList[bs].planeParams.pitch[i] * surface->surfaceList[bs].planeParams.height[i]);
        fwrite(surface->surfaceList[bs].mappedAddr.addr[i], bytes_to_write, 1, f);
        NvBufSurfaceUnMap(surface, bs, i);
      }
      fclose (f);
    } else {
      printf("error opening file for raw image writing\n");
    }
  }
}

NvBufSurface* getNvBufFromProbeInfo(GstPadProbeInfo *info) {
  if (!info) {
    GST_ERROR("GstPadProbeInfo null !");
    return NULL;
  }

  GstMapInfo map_info;
  GstBuffer* buffer = GST_PAD_PROBE_INFO_BUFFER (info);
  if (!gst_buffer_map(buffer, &map_info, GST_MAP_READ)) {
    GST_ERROR("gst_buffer_map() error!");
    return NULL;
  }
  NvBufSurface *surf = (NvBufSurface *) map_info.data;
  gst_buffer_unmap (buffer, &map_info);

  return surf;
}

/**
 * dec Buffer probe function on element.
 */
static GstPadProbeReturn
dec_get_raw_data_probe (GstPad * pad, GstPadProbeInfo *info, gpointer u_data)
{
  if (dec_probe_counter++ >= DUMP_FRAME_CNT_START &&
    dec_probe_counter <= DUMP_FRAME_CNT_STOP) {
    NvBufSurface* surface = getNvBufFromProbeInfo(info);

    int width = surface->surfaceList[0].width;
    int height = surface->surfaceList[0].height;
    int pitch = surface->surfaceList[0].pitch;
    int batchSize = surface->batchSize;
    int plane = surface->surfaceList[0].planeParams.num_planes;
    printf("dec width %d height %d pitch %d batchSize %d plane %d\n", width, height, pitch, batchSize, plane);
    dumpNvBufToFile("dec", "dec.data", dec_probe_counter, surface);
  }
  return GST_PAD_PROBE_OK;
}

/**
 * videoconvert Buffer probe function on element.
 */
static GstPadProbeReturn
nvvideoconvert_get_raw_data_probe (GstPad *pad, GstPadProbeInfo *info, gpointer u_data)
{
  if (nvvidcov_probe_counter++ >= DUMP_FRAME_CNT_START &&
    nvvidcov_probe_counter <= DUMP_FRAME_CNT_STOP) {
    NvBufSurface* surface = getNvBufFromProbeInfo(info);

    int width = surface->surfaceList[0].width;
    int height = surface->surfaceList[0].height;
    int pitch = surface->surfaceList[0].pitch;
    int batchSize = surface->batchSize;
    int plane = surface->surfaceList[0].planeParams.num_planes;
    printf("nvvideoconvert width %d height %d pitch %d batchSize %d plane %d\n", width, height, pitch, batchSize, plane);
    dumpNvBufToFile("nvvideoconvert", "nvvideoconvert.data", nvvidcov_probe_counter, surface);
  }
  return GST_PAD_PROBE_OK;
}

/**
 * streammux Buffer probe function on element.
 */
static GstPadProbeReturn
streammux_get_raw_data_probe (GstPad *pad, GstPadProbeInfo *info, gpointer u_data)
{
  if (streammux_probe_counter++ >= DUMP_FRAME_CNT_START &&
    streammux_probe_counter <= DUMP_FRAME_CNT_STOP) {
    NvBufSurface* surface = getNvBufFromProbeInfo(info);

    int width = surface->surfaceList[0].width;
    int height = surface->surfaceList[0].height;
    int pitch = surface->surfaceList[0].pitch;
    int batchSize = surface->batchSize;
    int plane = surface->surfaceList[0].planeParams.num_planes;
    printf("streammux width %d height %d pitch %d batchSize %d plane %d\n", width, height, pitch, batchSize, plane);
    dumpNvBufToFile("streammux", "streammux.data", streammux_probe_counter, surface);
  }

  return GST_PAD_PROBE_OK;
}

static gboolean
bus_call (GstBus * bus, GstMessage * msg, gpointer data)
{
  GMainLoop *loop = (GMainLoop *) data;
  switch (GST_MESSAGE_TYPE (msg)) {
    case GST_MESSAGE_EOS:
      g_print ("End of stream\n");
      g_main_loop_quit (loop);
      break;
    case GST_MESSAGE_WARNING:
    {
      gchar *debug;
      GError *error;
      gst_message_parse_warning (msg, &error, &debug);
      g_printerr ("WARNING from element %s: %s\n",
          GST_OBJECT_NAME (msg->src), error->message);
      g_free (debug);
      g_printerr ("Warning: %s\n", error->message);
      g_error_free (error);
      break;
    }
    case GST_MESSAGE_ERROR:
    {
      gchar *debug;
      GError *error;
      gst_message_parse_error (msg, &error, &debug);
      g_printerr ("ERROR from element %s: %s\n",
          GST_OBJECT_NAME (msg->src), error->message);
      if (debug)
        g_printerr ("Error details: %s\n", debug);
      g_free (debug);
      g_error_free (error);
      g_main_loop_quit (loop);
      break;
    }
#ifndef PLATFORM_TEGRA
    case GST_MESSAGE_ELEMENT:
    {
      if (gst_nvmessage_is_stream_eos (msg)) {
        guint stream_id;
        if (gst_nvmessage_parse_stream_eos (msg, &stream_id)) {
          g_print ("Got EOS from stream %d\n", stream_id);
        }
      }
      break;
    }
#endif
    default:
      break;
  }
  return TRUE;
}

static GstElement *
create_source_bin (guint index, gchar * uri)
{
  GstElement *bin = NULL/*, *uri_decode_bin = NULL*/;
  gchar bin_name[16] = { };
  gboolean multi_file_src = FALSE;

  int current_device = -1;
  cudaGetDevice(&current_device);
  struct cudaDeviceProp prop;
  cudaGetDeviceProperties(&prop, current_device);

  g_snprintf (bin_name, 15, "source-bin-%02d", index);
  /* Create a source GstBin to abstract this bin's content from the rest of the
   * pipeline */
  bin = gst_bin_new (bin_name);

  GstElement *source, *jpegparser, *decoder;

  if (strstr(uri, "%d"))
  {
    source = gst_element_factory_make("multifilesrc", "source");
    multi_file_src = TRUE;
  }
  else
    source = gst_element_factory_make ("filesrc", "source");

  jpegparser = gst_element_factory_make ("jpegparse", "jpeg-parser");

  decoder = gst_element_factory_make ("nvv4l2decoder", "nvv4l2-decoder");

  if (!source || !jpegparser || !decoder)
  {
    g_printerr ("One element could not be created. Exiting.\n");
    return NULL;
  }
    
  g_object_set (G_OBJECT (source), "location", uri, NULL);

  const char *dot = strrchr(uri, '.');
  if ((!strcmp (dot+1, "mjpeg")) || (!strcmp (dot+1, "mjpg")) || (!strcmp (dot+1, "mp4")) || (multi_file_src == TRUE))
  {
    if(prop.integrated) {
      g_object_set (G_OBJECT (decoder), "mjpeg", 1, NULL);
    }
  }

  gst_bin_add_many (GST_BIN (bin), source, jpegparser, decoder, NULL);

  gst_element_link_many (source, jpegparser, decoder, NULL);

  /* We need to create a ghost pad for the source bin which will act as a proxy
   * for the video decoder src pad. The ghost pad will not have a target right
   * now. Once the decode bin creates the video decoder and generates the
   * cb_newpad callback, we will set the ghost pad target to the video decoder
   * src pad. */
  if (!gst_element_add_pad (bin, gst_ghost_pad_new_no_target ("src",
              GST_PAD_SRC))) {
    g_printerr ("Failed to add ghost pad in source bin\n");
    return NULL;
  }

  GstPad *srcpad = gst_element_get_static_pad (decoder, "src");
  if (!srcpad) {
    g_printerr ("Failed to get src pad of source bin. Exiting.\n");
    return NULL;
  } else {
    gst_pad_add_probe(srcpad, GST_PAD_PROBE_TYPE_BUFFER,
        dec_get_raw_data_probe, NULL, NULL);
  }

  GstPad *bin_ghost_pad = gst_element_get_static_pad (bin, "src");
  if (!gst_ghost_pad_set_target (GST_GHOST_PAD (bin_ghost_pad),
        srcpad)) {
    g_printerr ("Failed to link decoder src pad to source bin ghost pad\n");
  }

  return bin;
}

int
main (int argc, char *argv[])
{
  GMainLoop *loop = NULL;
  GstElement *pipeline = NULL, *streammux = NULL, *sink = NULL, *nvvideoconv = NULL, *pgie = NULL,
      *nvvidconv = NULL, *nvosd = NULL, *tiler = NULL;
  GstBus *bus = NULL;
  guint bus_watch_id;
  GstPad *tiler_src_pad = NULL;
  guint i, num_sources;
  guint tiler_rows, tiler_columns;
  guint pgie_batch_size;

  int current_device = -1;
  cudaGetDevice(&current_device);
  struct cudaDeviceProp prop;
  cudaGetDeviceProperties(&prop, current_device);

  /* Check input arguments */
  if (argc < 2) {
    g_printerr ("Usage: %s <elementary JPEG file1> <elementary JPEG file2> ..."
        " <elementary JPEG fileN> \n", argv[0]);
    return -1;
  }
  num_sources = argc - 1;

  /* Standard GStreamer initialization */
  gst_init (&argc, &argv);
  loop = g_main_loop_new (NULL, FALSE);

  /* Create gstreamer elements */
  /* Create Pipeline element that will form a connection of other elements */
  pipeline = gst_pipeline_new ("dstest-image-decode-pipeline");

  /* Create nvstreammux instance to form batches from one or more sources. */
  streammux = gst_element_factory_make ("nvstreammux", "stream-muxer");

  if (!pipeline || !streammux) {
    g_printerr ("One element could not be created. Exiting.\n");
    return -1;
  }
  gst_bin_add (GST_BIN (pipeline), streammux);

  for (i = 0; i < num_sources; i++) {
    GstPad *sinkpad, *srcpad;
    gchar pad_name[16] = { };
    GstElement *source_bin = create_source_bin (i, argv[i + 1]);

    if (!source_bin) {
      g_printerr ("Failed to create source bin. Exiting.\n");
      return -1;
    }

    gst_bin_add (GST_BIN (pipeline), source_bin);

    g_snprintf (pad_name, 15, "sink_%u", i);
    sinkpad = gst_element_get_request_pad (streammux, pad_name);
    if (!sinkpad) {
      g_printerr ("Streammux request sink pad failed. Exiting.\n");
      return -1;
    }

    srcpad = gst_element_get_static_pad (source_bin, "src");
    if (!srcpad) {
      g_printerr ("Failed to get src pad of source bin. Exiting.\n");
      return -1;
    }

    if (gst_pad_link (srcpad, sinkpad) != GST_PAD_LINK_OK) {
      g_printerr ("Failed to link source bin to stream muxer. Exiting.\n");
      return -1;
    }

    gst_object_unref (srcpad);
    gst_object_unref (sinkpad);
  }

  /* Use convertor to convert to appropriate format */
  nvvideoconv = gst_element_factory_make ("nvvideoconvert", "nvvideo-converter1");

  /* Use nvinfer to infer on batched frame. */
  pgie = gst_element_factory_make ("nvinfer", "primary-nvinference-engine");

  /* Use nvtiler to composite the batched frames into a 2D tiled array based
   * on the source of the frames. */
  tiler = gst_element_factory_make ("nvmultistreamtiler", "nvtiler");

  /* Use convertor to convert from NV12 to RGBA as required by nvosd */
  nvvidconv = gst_element_factory_make ("nvvideoconvert", "nvvideo-converter");

  /* Create OSD to draw on the converted RGBA buffer */
  nvosd = gst_element_factory_make ("nvdsosd", "nv-onscreendisplay");

  /* Finally render the osd output */
  if(prop.integrated) {
    sink= gst_element_factory_make ("nv3dsink", "nvvideo-renderer");
  } else {
    sink = gst_element_factory_make ("nveglglessink", "nvvideo-renderer");
  }
  g_object_set (G_OBJECT (sink), "sync", 0, NULL);

  if (!nvvideoconv || !pgie || !tiler || !nvvidconv || !nvosd || !sink) {
    g_printerr ("One element could not be created. Exiting.\n");
    return -1;
  }

  g_object_set (G_OBJECT (streammux), "width", MUXER_OUTPUT_WIDTH, "height",
      MUXER_OUTPUT_HEIGHT, "batch-size", num_sources,
      "batched-push-timeout", MUXER_BATCH_TIMEOUT_USEC, NULL);

  /* Configure the nvinfer element using the nvinfer config file. */
  g_object_set (G_OBJECT (pgie),
      "config-file-path", "dstest_image_decode_pgie_config.txt", NULL);

  /* Override the batch-size set in the config file with the number of sources. */
  g_object_get (G_OBJECT (pgie), "batch-size", &pgie_batch_size, NULL);
  if (pgie_batch_size != num_sources) {
    g_printerr
        ("WARNING: Overriding infer-config batch-size (%d) with number of sources (%d)\n",
        pgie_batch_size, num_sources);
    g_object_set (G_OBJECT (pgie), "batch-size", num_sources, NULL);
  }

  tiler_rows = (guint) sqrt (num_sources);
  tiler_columns = (guint) ceil (1.0 * num_sources / tiler_rows);
  /* we set the tiler properties here */
  g_object_set (G_OBJECT (tiler), "rows", tiler_rows, "columns", tiler_columns,
      "width", TILED_OUTPUT_WIDTH, "height", TILED_OUTPUT_HEIGHT, NULL);

  /* we add a message handler */
  bus = gst_pipeline_get_bus (GST_PIPELINE (pipeline));
  bus_watch_id = gst_bus_add_watch (bus, bus_call, loop);
  gst_object_unref (bus);

  /* Set up the pipeline */
  /* we add all elements into the pipeline */
  gst_bin_add_many (GST_BIN (pipeline), nvvideoconv, pgie, tiler, nvvidconv, nvosd, sink,
      NULL);
  /* we link the elements together
  * nvstreammux -> nvvideoconv -> nvinfer -> nvtiler -> nvvidconv -> nvosd -> video-renderer */
  if (!gst_element_link_many (streammux, nvvideoconv, pgie, tiler, nvvidconv, nvosd, sink,
          NULL)) {
    g_printerr ("Elements could not be linked. Exiting.\n");
    return -1;
  }
 
  GstPad *streammux_src_pad = gst_element_get_static_pad (streammux, "src");
  if (!streammux_src_pad)
    GST_ERROR ("Unable to get streammux_src_pad\n");
  else
    gst_pad_add_probe(streammux_src_pad, GST_PAD_PROBE_TYPE_BUFFER,
        streammux_get_raw_data_probe, NULL, NULL);
  gst_object_unref (streammux_src_pad);

  GstPad *nvvideoconv_src_pad = gst_element_get_static_pad (nvvideoconv, "src");
  if (!nvvideoconv_src_pad)
    GST_ERROR ("Unable to get nvvideoconv_src_pad\n");
  else
    gst_pad_add_probe(nvvideoconv_src_pad, GST_PAD_PROBE_TYPE_BUFFER,
        nvvideoconvert_get_raw_data_probe, NULL, NULL);
  gst_object_unref (nvvideoconv_src_pad);

  /* Lets add probe to get informed of the meta data generated, we add probe to
   * the sink pad of the osd element, since by that time, the buffer would have
   * had got all the metadata. */
  tiler_src_pad = gst_element_get_static_pad (pgie, "src");
  if (!tiler_src_pad)
    g_print ("Unable to get src pad\n");
  else
    gst_pad_add_probe (tiler_src_pad, GST_PAD_PROBE_TYPE_BUFFER,
        tiler_src_pad_buffer_probe, NULL, NULL);
  gst_object_unref (tiler_src_pad);

  /* Set the pipeline to "playing" state */
  g_print ("Now playing:");
  for (i = 0; i < num_sources; i++) {
    g_print (" %s,", argv[i + 1]);
  }
  g_print ("\n");
  gst_element_set_state (pipeline, GST_STATE_PLAYING);

  /* Wait till pipeline encounters an error or EOS */
  g_print ("Running...\n");
  g_main_loop_run (loop);

  /* Out of the main loop, clean up nicely */
  g_print ("Returned, stopping playback\n");
  gst_element_set_state (pipeline, GST_STATE_NULL);
  g_print ("Deleting pipeline\n");
  gst_object_unref (GST_OBJECT (pipeline));
  g_source_remove (bus_watch_id);
  g_main_loop_unref (loop);
  return 0;
}
